/* GNU polyxmass - the massist's program.
   -------------------------------------- 
   Copyright (C) 2000,2001,2002 Filippo Rusconi

   http://www.polyxmass.org

   This file is part of the "GNU polyxmass" project.
   
   The "GNU polyxmass" project is an official GNU project package (see
   www.gnu.org) released ---in its entirety--- under the GNU General
   Public License and was started at the Centre National de la
   Recherche Scientifique (FRANCE), that granted me the formal
   authorization to publish it under this Free Software License.

   This software is free software; you can redistribute it and/or
   modify it under the terms of the GNU  General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU  General Public
   License along with this software; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include "pxmutils-config.h"
#include "pxmutils-globals.h"


/* Below, CONFDIR is defined in configure.in to be 
   "${packageprefix}/etc/polyxmass.d", thus if $(prefix) is 
   /usr, for example, config.h will contain a line saying:
   #define CONFDIR "/usr/etc/polyxmass.d"
*/
gchar etc_polyxmassdata_cfg_file [] = \
CONFDIR"/polyxmassdata.conf";


/* Returns the directory name where the default polyxmassdata
   are. That is an allocated string. Must be freed when no more in
   use.
*/
gchar *
pxmutils_config_get_sys_polyxmassdata_dir (void)
{
  gchar *read = NULL;
  gchar *line = NULL;
  gchar *found = NULL;
  
  FILE *cfgp = NULL;

  /* This function checks that the $(prefix)/etc/polyxmass.d directory
     exists and that in it the file polyxmassdata.conf exists,
     containining a valid directory name.
  */
  if (FALSE == g_file_test (etc_polyxmassdata_cfg_file, G_FILE_TEST_EXISTS))
    return FALSE;

  cfgp = fopen (etc_polyxmassdata_cfg_file, "r");

  if (cfgp == NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: error with opening file '%s'\n"),
	     __FILE__, __LINE__, etc_polyxmassdata_cfg_file);
      
      return NULL;
    }
  
  /* Allocate the string (empty) where the lines read from the file
     are to be stored for later analysis.
  */
  line = g_malloc0 (sizeof (gchar) * (MAX_LINE_LENGTH + 1));

  /* Get the single line from the file.
   */
  read = line;
  
  while (read != NULL)
    {
      read = fgets (read, MAX_LINE_LENGTH, cfgp);
      
      if (read != NULL)
	{
	  if (strlen (read) > MAX_LINE_LENGTH)
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		     _("%s@%d: the line read from file cannot be larger"
		       " than the authorized length! This is dangerous.\n"),
		     __FILE__, __LINE__);
	    }
	  
	  found = strstr (read, "polyxmassdata=");
	  
	  if (found == NULL)
	    continue;
	  else
	    break;
	}
    }
  
  if (found != NULL)
    {
      /* There was a line holding something like "main_config_dir=". We
	 now need to get a pointer from '=' to end of line.
      */
      found =  strchr (found, '=');
      /* found is now "=/any_directory/any_file". So we need to go one 
	 character further.
      */
      found++;
      
      /* Make an allocated copy of the string so that we can unspacify it.
       */
      found = g_strdup (found);
      
      found = pxmutils_globals_unspacify_string (&found);
      
      if (FALSE == g_file_test (found, G_FILE_TEST_IS_DIR))
	{
	  g_free (line);
	  g_free (found);
	  
	  fclose (cfgp);
	  
	  return NULL;
	}
      else
	{
	  g_free (line);
	  
	  fclose (cfgp);
	  
	  return found;
	}
    }
    
  g_free (line);
  
  fclose (cfgp);
  
  return NULL;
}


/* Returns the directory name where the user's (or default)
   polyxmassdata configuration data are. That is an allocated
   string. Must be freed when no more in use.
*/
gchar *
pxmutils_config_get_user_polyxmassdata_dir (gchar *username, 
					    gchar *homedir)
{
  gchar *help;
  gchar *dir;
  
  gchar *read = NULL;
  gchar *line = NULL;
  gchar *found = NULL;
  
  FILE *cfgp = NULL;


  /* This function checks that the the .polyxmass.d directory exists
     and that polyxmassdata.conf exists in it. If any of them does not
     exist, it is created and $(prefix)/etc/polyxmassdata.conf is
     copied.  Note that username can be NULL, in which case homedir
     cannot, and vice versa.
  */
  g_assert (username != NULL || homedir != NULL);
  
  
  if (username != NULL)
    {
      help = (gchar *) g_get_home_dir () ; /* not allocated ! */
      dir = g_strdup_printf ("%s/.polyxmass.d", help);
    }
  else
    {
      /* Make sure that homedir is indeed a directory.
       */
      if (FALSE == g_path_is_absolute (homedir))
	return NULL;
      
      if (FALSE == g_file_test (homedir, G_FILE_TEST_IS_DIR))
	return NULL;
      
      dir = g_strdup_printf ("%s/.polyxmass.d", homedir);
    }

  /* We now have a directory name which is valid. We have to check
     that it exists indeed. Otherwise we just create it.
   */
  if (FALSE == g_file_test (dir, G_FILE_TEST_IS_DIR))
    {
      /* We create it !
       */
      if (0 != mkdir (dir, S_IRWXU))
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		 _("%s@%d: failed creating directory '%s' "
		   "- Error: '%s'\n"),
		 __FILE__, __LINE__, dir, strerror (errno));
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
		 _("%s@%d: created directory '%s'\n"),
		 __FILE__, __LINE__, dir);
	}
    }
  
  /* Now check that the file exists in this directory. Make the file name
     first.
   */
  help = g_strdup_printf ("%s/polyxmassdata.conf", dir);
  
  /* Free 'dir', do not need it anymore.
   */
  g_free (dir);
  
  if (FALSE == g_file_test (help, G_FILE_TEST_EXISTS))
    {
      /* The file does not exist, which means we just copy it from
	 "/etc/polyxmass.d/polyxmassdata.conf".
      */
      if (FALSE == pxmutils_globals_copy_files (help, 
						etc_polyxmassdata_cfg_file))
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		 _("%s@%d: failed copying file '%s' to '%s' "
		   "- Error: '%s'\n"),
		 __FILE__, __LINE__, 
		 etc_polyxmassdata_cfg_file,
		 help,
		 strerror (errno));
	}
      else
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE,
		 _("%s@%d: copied file '%s' to '%s'\n"),
		 __FILE__, __LINE__, etc_polyxmassdata_cfg_file, help);
	}
    }

  /* Now that we have the absolute name of the user's
     polyxmassdata.conf file, we can get its contents, check that they
     point to an existent polyxmassdata directory.
  */
  cfgp = fopen (help, "r");

  if (cfgp == NULL)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: error with opening file '%s'\n"),
	     __FILE__, __LINE__, etc_polyxmassdata_cfg_file);
      
      g_free (help);
      
      return NULL;
    }
  
  g_free (help);

  /* Get the single line from the file.
   */
  line = g_malloc0 (sizeof (gchar) * (MAX_LINE_LENGTH + 1));
  read = line;
  
  while (read != NULL)
    {
      read = fgets (read, MAX_LINE_LENGTH, cfgp);
      debug_printf (("read string is %s\n", read));
      
      if (read != NULL)
	{
	  if (strlen (read) > MAX_LINE_LENGTH)
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		   _("%s@%d: the line read from file cannot be larger"
		     " than the authorized length! This is dangerous.\n"),
		   __FILE__, __LINE__);

	  found = strstr (read, "polyxmassdata=");

	  if (found == NULL)
	    continue;
	  else
	    break;
	}
    }
  
  if (found != NULL)
    {
      /* There was a line holding something like "main_config_dir=". We
	 now need to get a pointer from '=' to end of line.
      */
      found = strchr (found, '=');
      /* found is now "=/any_directory/any_file". So we need to go one 
	 character further.
      */
      found++;
      
      /* Make an allocated copy of the string so that we can unspacify it.
       */
      found = g_strdup (found);
      
      found = pxmutils_globals_unspacify_string (&found);
      
      if (FALSE == g_file_test (found, G_FILE_TEST_IS_DIR))
	{
	  g_free (line);
	  g_free (found);

	  fclose (cfgp);
	  
	  return NULL;
	}
      else
	{
	  g_free (line);
	  fclose (cfgp);
	  
	  return found;
	}
    }
    
  fclose (cfgp);

  return found;
}

			    
