/* GNU polyxmass - the massist's program.
   -------------------------------------- 
   Copyright (C) 2000,2001,2002 Filippo Rusconi

   http://www.polyxmass.org

   This file is part of the "GNU polyxmass" project.
   
   The "GNU polyxmass" project is an official GNU project package (see
   www.gnu.org) released ---in its entirety--- under the GNU General
   Public License and was started at the Centre National de la
   Recherche Scientifique (FRANCE), that granted me the formal
   authorization to publish it under this Free Software License.

   This software is free software; you can redistribute it and/or
   modify it under the terms of the GNU  General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU  General Public
   License along with this software; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "pxmutils-note.h"



/* NEW'ING FUNCTIONS, DUPLICATING FUNCTIONS ...
 */
PxmNote*
pxmutils_note_new (void)
{
  PxmNote *note = g_malloc0 (sizeof (PxmNote));

  note->notevalGPA = g_ptr_array_new ();

  return note;
}

PxmNoteVal*
pxmutils_noteval_new (void)
{
  PxmNoteVal *noteval = g_malloc0 (sizeof (PxmNoteVal));

  return noteval;
}


PxmProp*
pxmutils_note_prop_new (void)
{
  PxmProp *prop = pxmutils_prop_new ();
  
  PxmNote *note = pxmutils_note_new ();

  pxmutils_prop_set_name (prop, "NOTE");
  
  prop->data = (gpointer) note;

  prop->custom_dup = pxmutils_note_prop_dup;
  prop->custom_cmp = pxmutils_note_prop_cmp;
  prop->custom_free = pxmutils_note_prop_free;
  
  return prop;
}


PxmNoteVal*
pxmutils_noteval_dup (PxmNoteVal *noteval)
{
  PxmNoteVal *noteval_new = NULL;
  
  g_assert (noteval != NULL);
  
  noteval_new = pxmutils_noteval_new ();
  
  if (noteval->value != NULL)
    {
      if (noteval->type == PXM_NOTE_VAL_TYPE_STR)
	noteval_new->value = (gpointer) g_strdup ((gchar *) noteval->value);
      else if (noteval->type == PXM_NOTE_VAL_TYPE_INT)
	{
	  noteval_new->value = g_malloc0 (sizeof (gint));
	  (*(gint *) noteval_new->value = *(gint *) noteval->value);
	}
      else if (noteval->type == PXM_NOTE_VAL_TYPE_DBL)
	{
	  noteval_new->value = g_malloc0 (sizeof (gdouble));
	  (*(gdouble *) noteval_new->value = *(gdouble *) noteval->value);
	}
      else
	g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
	       _("%s@%d: bad type for noteval value. Exiting.\n"),
	       __FILE__, __LINE__);
    }

  noteval_new->type = noteval->type;
      
  return noteval_new;
}


PxmNote*
pxmutils_note_dup (PxmNote *note)
{
  PxmNote *note_new = NULL;
  PxmNoteVal *noteval = NULL;
  PxmNoteVal *noteval_new = NULL;
  
  gint iter = 0;
  
  g_assert (note != NULL);
  
  note_new = pxmutils_note_new ();
  
  g_assert (note->name != NULL);
  note_new->name = g_strdup (note->name);
  
  g_assert (note->notevalGPA != NULL);
  for (iter = 0; iter < note->notevalGPA->len ; iter ++)
    {
      noteval = g_ptr_array_index (note->notevalGPA, iter);
      g_assert (noteval != NULL);
      
      noteval_new = pxmutils_noteval_dup (noteval);
      
      g_ptr_array_add (note_new->notevalGPA, noteval_new);
    }
  
  return note_new;
}


PxmProp*
pxmutils_note_prop_dup (PxmProp *prop)
{
  PxmProp *prop_new = NULL;
  PxmNote *note_new = NULL;

  g_assert (prop != NULL);
  
  g_assert (prop->name != NULL && 0 == strcmp ("NOTE", prop->name));
  g_assert (prop->data != NULL);
  

  prop_new = pxmutils_prop_new ();
  
  pxmutils_prop_set_name (prop_new, "NOTE");
  
  note_new = pxmutils_note_dup ((PxmNote *) prop->data);

  prop_new->data = (gpointer) note_new;

  prop_new->custom_dup = prop->custom_dup;
  prop_new->custom_cmp = prop->custom_cmp;
  prop_new->custom_free = prop->custom_free;
    
  return prop_new;
}


/*  LOCATING FUNCTIONS
 */
gint
pxmutils_noteval_get_index_by_noteval (PxmNoteVal *noteval, GPtrArray *GPA)
{
  gint iter = 0;

  PxmNoteVal *noteval_iter = NULL;
  
  /* We have to localize a noteval object in GPA that is identical to
     the one passed as parameter.
  */
  g_assert (noteval != NULL);
  g_assert (GPA != NULL);
  
  for (iter = 0; iter < GPA->len; iter++)
    {
      noteval_iter = g_ptr_array_index (GPA, iter);
      g_assert (noteval_iter != NULL);

      if (0 == pxmutils_noteval_cmp (noteval, noteval_iter))
	return iter;
    }
  
  return -1;
}



/* DATA MODIFYING FUNCTIONS
 */
gchar* 
pxmutils_note_set_name (PxmNote *note, gchar *name)
{
  g_assert (note != NULL && name != NULL);

  /* If name already allocated, free it before.
   */
  if (note->name != NULL)
    g_free (note->name);

  note->name = g_strdup (name);

  return note->name;
}



/* COMPARISON FUNCTIONS
 */
gint
pxmutils_noteval_cmp (PxmNoteVal *noteval1,
		      PxmNoteVal *noteval2)
{
  g_assert (noteval1 != NULL && noteval2 != NULL);
  
  /* On the ground that other similar functions return 0 if the
     operands are identical, we will do the same. Return 1 if the
     operands are different.
  */
  if (noteval1->type != noteval2->type)
    return 1;
  
  if (noteval1->value == NULL && noteval2->value == NULL)
    return 0;
  
  if (noteval1->value != NULL && noteval2->value == NULL)
    return 1;

  if (noteval1->value == NULL && noteval2->value != NULL)
    return 1;

  /* At this point we know that both pointers are not NULL, which
     means we will not crash using them to perform comparisons. Note
     that the functions used to compare the data pointed to by the
     pointers are different depending on the type of the pointers !
  */
  if (noteval1->type == PXM_NOTE_VAL_TYPE_STR)
    return strcmp ((gchar *) noteval1->value, (gchar *) noteval2->value);
  else if (noteval1->type == PXM_NOTE_VAL_TYPE_INT)
    return (*(gint *) noteval1->value != *(gint *) noteval2->value);
  else if (noteval1->type == PXM_NOTE_VAL_TYPE_DBL)
    return (*(gdouble *) noteval1->value != *(gdouble *) noteval2->value);
  else
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
	   _("%s@%d: bad type for noteval values. Exiting.\n"),
	   __FILE__, __LINE__);

  /* Do not care what we return, because of the error with abortion
     right above here.
  */
  return 1;
}


gint
pxmutils_note_cmp (gint what /*NOTE_NAME|NOTE_VALUE|NOTE_BOTH*/,
		   PxmNote *note1,
		   PxmNote *note2)
{
  /* Since both members of the PxmNote object are strings, they are
   * compared using strcmp (). The returned value is not 0 if
   * differences are encountered.
   */
  gint resultName = 0;
  gint resultValue = 0;
  gint iter = 0;

  PxmNoteVal *noteval1 = NULL;
  PxmNoteVal *noteval2 = NULL;
 

  g_assert (note1 != NULL && note2 != NULL);


  if (what == NOTE_NAME || what == NOTE_BOTH)
    {
      g_assert (note1->name != NULL && note2->name != NULL);

      resultName = strcmp (note1->name, note2->name);
      
      if (what == NOTE_NAME)
	return resultName;
    }

  if (what == NOTE_VALUE || what == NOTE_BOTH)
    {
      /* What we have to actually compare is the arrays of PxmNoteVal
	 objects. First off, clearly, if the arrays are not of the
	 same length, that means that they are different! Next,
	 compare one by one all the PxmNoteVal objects in the arrays.
      */
      g_assert (note1->notevalGPA != NULL && note2->notevalGPA != NULL);

      if (note1->notevalGPA->len != note2->notevalGPA->len)
	resultValue = 1;
      else
	{
	  for (iter = 0 ; iter < note1->notevalGPA->len ; iter++)
	    {
	      noteval1 = g_ptr_array_index (note1->notevalGPA, iter);
	      g_assert (noteval1 != NULL);

	      noteval2 = g_ptr_array_index (note2->notevalGPA, iter);
	      g_assert (noteval2 != NULL);
	      
	      /* The function call below will take care of comparing the
		 two noteval objects according to the type of the data.
	      */
	      resultValue = pxmutils_noteval_cmp (noteval1, noteval2);
	      if (resultValue != 0)
		break;
	    }
	}
      
      if (what == NOTE_VALUE)
	return resultValue;
    }

  if (what == NOTE_BOTH)
    return (resultName + resultValue);

  else
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
	   _("%s@%d: the what param is not valid. Exiting.\n"),
	   __FILE__, __LINE__);

  return 1;
}


gint
pxmutils_note_prop_cmp (gint what /*PROP_NAME|PROP_DATA|PROP_BOTH*/,
			PxmProp *prop1,
			PxmProp *prop2)
{
  /* This function is called when two PxmProp instances named
     "NOTE" must be compared.
  */
  PxmNote *note1 = NULL;
  PxmNote *note2 = NULL;
  
  gint resultName = 0;
  gint resultData = 0;

  g_assert (prop1 != NULL && prop2 != NULL);


  if (what == PROP_NAME || what == PROP_BOTH)
    {
      g_assert (prop1->name != NULL && prop2->name != NULL );

      resultName = strcmp (prop1->name, prop2->name);
      
      if (what == PROP_NAME)
	return resultName;
    }

  if (what == PROP_DATA || what == PROP_BOTH)
    {
      /* We have come to the PxmNote-specific point. We must compare
	 both the name and value members of the PxmNote instance.
       */
      note1 = (PxmNote *) prop1->data;
      note2 = (PxmNote *) prop2->data;
      
      g_assert (note1 != NULL && note2 != NULL);

      resultData = pxmutils_note_cmp (NOTE_BOTH, note1, note2);

      if (what == PROP_DATA)
	return resultData;
    }

  if (what == PROP_BOTH)
    return (resultName + resultData);

  else
    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
	   _("%s@%d: the what param is not valid. Exiting.\n"),
	   __FILE__, __LINE__);

  return 1;
}


gint
pxmutils_note_prop_data_cmp (gpointer data1,
			     gpointer data2)
{
  /* This function is called when two PxmNote objects are to be
     compared. These two note objects are initially the data member of
     PxmProp objects. Which is why we get gpointer parameters, that
     will need to be cast to PxmNote *.
  */
  PxmNote *note1 = NULL;
  PxmNote *note2 = NULL;
  
  note1 = (PxmNote *) data1;
  note2 = (PxmNote *) data2;
 
  g_assert (note1 != NULL && note2 != NULL);

  return pxmutils_note_cmp (NOTE_BOTH, note1, note2);

}


/* RENDERING FUNCTIONS (FORMATTER FUNCTION)
 */
gchar*
pxmutils_note_prop_format_xml_node_note (PxmProp *prop,
					 gchar *indent,
					 gint offset,
					 gpointer user_data)
{
  /* This formatter is used when a polymer sequence has property
     object (type PxmProp) of name "NOTE" (for which the data
     points to a PxmNote object). If such polymer sequence has to be
     saved to a file, under format xml, then when the saving function
     encounter a "NOTE"-named PxmProp object, it calls this
     saving function which is responsible for creating a character
     string to be sent to the xml file.

     The name of a PxmNote is a string of character, which is saved
     very easily in the xml file, under element <name>. The PxmNote
     object contains an array of PxmNoteval objects, which have to be
     deconstructed one by one (see array-iterating loop below) into a
     <data> element with an attribute that can either be "str" or
     "int" or "dbl", respectively to indicate that the data in the
     PxmNoteval object is text or numerical gint or gdouble.
  */
  PxmNote *note = NULL;
  
  GString *gs = NULL;
  
  gchar *lead = NULL;
  gchar *help = NULL;
  PxmNoteVal *noteval = NULL;
  
  gint new_offset = 0;
  gint iter = 0;
  
  g_assert (prop != NULL);
  g_assert (prop->name != NULL && prop->data != NULL);
  
  /* OK, the property to be formatter as XML text is not NULL and none
     of its member either. Let's get to the PxmNote object that is
     pointed by the prop->data member.
  */
  note = (PxmNote *) prop->data;
  g_assert (note->name != NULL && note->notevalGPA != NULL);
  
  /* OK, both the name and value members of the PxmNote object
     are non-NULL. We can start doing our work.
  */
  
  gs = g_string_new ("");
  g_assert (gs != NULL);

  /* We are willing to create a SINGLE <note> node that should look 
     like this:

     <prop>
       <name>NOTE</name>
       <data type="str" "WHATEVER_STRING: the note's name</data>
       <data type="int" WHATEVER_STRING: the note's name</data>
       <data type="dbl" WHATEVER_STRING: the note's value</data>
     </prop>
     
  */

  new_offset = offset;

  /* Open the <prop> element.
   */
  lead = pxmutils_globals_format_string_lead (indent, new_offset);

  g_string_append_printf (gs, "%s<prop>\n", lead);

  g_free (lead);
  new_offset++;
  lead = pxmutils_globals_format_string_lead (indent, new_offset);
  
  /* The prop name member.
   */
  g_string_append_printf (gs, "%s<name>%s</name>\n",
			  lead, prop->name);
  
  /* The note name member.
   */
  g_string_append_printf (gs, "%s<data>%s</data>\n",
			  lead, (gchar*) note->name);
  
  /* We now have to iterate in the note->valGPA and for each string
     found in the array we have to make a <data> element.
  */
  for (iter = 0 ; iter < note->notevalGPA->len ; iter++)
    {
      noteval = g_ptr_array_index (note->notevalGPA, iter);
      g_assert (noteval != NULL);
      
      /* Make the string to append to the gs xml string. The way the string
	 is made depends on the type of the value:
       */
      if (noteval->type == PXM_NOTE_VAL_TYPE_STR)
	g_string_append_printf (gs, 
				"%s<data %s>%s</data>\n",
				lead, 
				"type=\"str\"",
				(gchar*) noteval->value);
      else if (noteval->type == PXM_NOTE_VAL_TYPE_INT)
	g_string_append_printf (gs, 
				"%s<data %s>%d</data>\n",
				lead, 
				"type=\"int\"",
				*(gint *) noteval->value);
      else if (noteval->type == PXM_NOTE_VAL_TYPE_DBL)
	g_string_append_printf (gs, 
				"%s<data %s>%.9f</data>\n",
				lead, 
				"type=\"dbl\"",
				*(gdouble *) noteval->value);
      else g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		  _("%s@%d: the noteval->type is not valid. Exiting.\n"),
		  __FILE__, __LINE__);
    }
  
  g_free (lead);
  new_offset--;
  lead = pxmutils_globals_format_string_lead (indent, new_offset);
  
  /* Close the note element.
   */
  g_string_append_printf (gs, "%s</prop>\n", lead);
  
  g_free (lead);
  
  help = gs->str;
  
  g_string_free (gs, FALSE);
  
  return help;
}


PxmProp*
pxmutils_note_prop_render_xml_node_note (xmlDocPtr xml_doc,
					 xmlNodePtr xml_node,
					 gpointer user_data)
{
  /* This xml renderer is responsible for rendering a xml note that
     corresponds to a <prop> element of which the <name> element has
     the contents "NOTE". Thus, the property that is being
     created has a name "NOTE", and its pointer is pointing to
     an allocated PxmNote object.
  */

  /* The xml node we are in is structured this way:

     <prop>
       <name>NOTE</name>
       <data>WHATEVER_STRING: the note's name</data>
       <data type="str">WHATEVER_STRING: the 1st note's value</data>
       <data type="dbl">WHATEVER_STRING: the 2nd note's value</data>
       ... n-1
       <data type="str">WHATEVER_STRING: the nst note's value</data>
     </prop>

     The xml_node param should point to the xml element <prop> which
    is the entry node to the xml data describing the prop object.
   
    That is, xml_node->name == "prop", as exemplified below:
   
    <prop>
     ^
     |
     +------- here we are right now.
      <name>NOTE</name>
    
    Which means that xml_node->name == "prop" and that following the
    call to xml_node->children we should be pointing to the <name>
    child node of the <prop> node, thus:
    xml_node->xmlChildrenNode->content == "NOTE".
  */
  PxmProp *prop = NULL;
  PxmNote *note = NULL;
  PxmNoteVal *noteval = NULL;

  gchar *value = NULL;
  gchar *type = NULL;

  g_assert (xml_node != NULL);
  g_assert (xml_doc != NULL);

  g_assert (0 == xmlStrcmp (xml_node->name, (const xmlChar *) "prop"));
  
  /* Allocate the PxmProp object, so that we'll fill it with read
     data.
   */
  prop = pxmutils_prop_new ();
  
  /* Go to the <name> node, which is the first <prop> element's child:
   */
  xml_node = xml_node->children;  

  /* From a rigourous XML parsing point of view, the blanks found in
     the XML document are considered to be nodes, and we have to
     detect these and take proper action: go next sibling (next blank) as
     long as blanks are encountered.
   */
  while (TRUE == xmlIsBlankNode(xml_node))
    xml_node = xml_node->next;

  /* We have encountered the next children's element: it should be
     <name>. That <name> element should contain the "NOTE" string.
   */
  g_assert (0 == strcmp (xml_node->name, (const xmlChar *) "name"));

  /* Get the name of the property so that we can start filling the
   * prop instance.
   */
  prop->name = 
    xmlNodeListGetString (xml_doc, xml_node->xmlChildrenNode, 1);
  g_assert (prop->name != NULL);
  g_assert (0 == strcmp ("NOTE", prop->name));

  /* We are now certain that we are effectively dealing with a NOTE 
     prop object, so create such PxmNote object now.
  */
  note = pxmutils_note_new ();

  /* Go to next child of the prop element, which, according to DTD
     should be one (or more) <data> element(s). To go to the <data>
     element just go next. In this xml renderer there are two <data>
     elements that contain #PCDATA, which means that we just allocate
     strings.
   */
  xml_node = xml_node->next;
  while (TRUE == xmlIsBlankNode(xml_node))
    xml_node = xml_node->next;

  /* We got to the first <data> element, which is going to be the 
     PxmNote note->name member.
  */
  g_assert (0 == strcmp (xml_node->name, "data"));

  note->name = xmlNodeListGetString (xml_doc, xml_node->xmlChildrenNode, 1);
  g_assert (note->name != NULL);
  
  /* At this point go to the element(s) that is (are) to become the
     note value(s). There may be more than one value, and we have to
     parse all the available values as strings. Each time we find a string
     we must add it to the note->valGPA array of value strings.
  */
  xml_node = xml_node->next;
  while (TRUE == xmlIsBlankNode(xml_node))
    xml_node = xml_node->next;

  while (xml_node != NULL)
    {
      /* If there is a node, it cannot be of any name other than "data".
       */
      g_assert (0 == strcmp (xml_node->name, "data"));

      /* Collect the value that is associated with the "data"
	 element. That string is to go to the note->valGPA array of
	 value strings. That string must not be NULL, otherwise the
	 existence _per se_ of the "data" element is questionable !
       */
      value = 
	xmlNodeListGetString (xml_doc, xml_node->xmlChildrenNode, 1);
      g_assert (value != NULL);

      /* We know that we are going to need a PxmNoteVal object. So just
	 create it;
      */
      noteval = pxmutils_noteval_new ();
      
      /* Get to know if there is an attribute (type) --we know it is
	 optional, if it is absent, then that equals to a string
	 type.
      */
      type = xmlGetProp (xml_node, "type");

      /*
	debug_printf (("type is now %s\n", type));
      */

      if (type == NULL)
	{
	  /* The value is by default of type string, so we just use
	     the value string obtained above to make the PxmNoteVal
	     object.
	  */
	  noteval->type = PXM_NOTE_VAL_TYPE_STR;
	  noteval->value = (gpointer) value;
	}
      else
	{
	  if (0 == strcmp ("str", type))
	    {
	      noteval->type = PXM_NOTE_VAL_TYPE_STR;
	      noteval->value = (gpointer) g_strdup (value);
	    }
	  else if (0 == strcmp ("int", type))
	    {
	      noteval->type = PXM_NOTE_VAL_TYPE_INT;
	      noteval->value = g_malloc0 (sizeof (gint));
	      if (FALSE == 
		  pxmutils_globals_strtoi (value,
					   (gint *) noteval->value,
					   10))
		g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		       _("%s@%d: error converting value to gint. "
			 "Exiting.\n"),
		       __FILE__, __LINE__);
	    }
	  else if (0 == strcmp ("dbl", type))
	    {
	      noteval->type = PXM_NOTE_VAL_TYPE_DBL;
	      noteval->value = g_malloc0 (sizeof (gdouble));
	      if (FALSE == 
		  pxmutils_globals_strtod (value,
					   (gdouble *) noteval->value))
		g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		       _("%s@%d: error converting value to gdouble. "
			 "Exiting.\n"),
		       __FILE__, __LINE__);
	    }
	  else
	    g_log (G_LOG_DOMAIN, G_LOG_LEVEL_ERROR,
		   _("%s@%d: file badly formed. Exiting.\n"),
		   __FILE__, __LINE__);

	  /* At this point we do not need the attribute anymore.
	   */
	  xmlFree (type);
      	}
      
      /* OK, now that we have our noteval object, we can just append
	 it to the note->notevalGPA:
       */
      g_ptr_array_add (note->notevalGPA, noteval);
      
      /* At this point, go the next "data" element, if any.
       */
      xml_node = xml_node->next;

      /* From a rigorous XML parsing point of view, the blanks found in
         the XML document are considered to be nodes, and we have to detect
         these and take proper action: go next sibling (next blank) as long
         as blanks are encountered.
      */
      while (TRUE == xmlIsBlankNode (xml_node))
	xml_node = xml_node->next;
    }

  /* At this point, the PxmNote object 'note' has been filled OK, and
     now we have to set its pointer to the data member of the PxmProp
     object 'prop'.
  */
  prop->data = (gpointer) note;
  
  /* At this point we have parsed the <prop> node, of name "NOTE"
     into a prop object of name "NOTE" and with its data pointing
     to a PxmNote object.
  */

  /* Do not forget that when this property is used, it might be
     either compared or duplicated or freed, and we must set
     to it the pointers to the functions that can accomplish those
     tasks properly.
  */

  prop->custom_dup = pxmutils_note_prop_dup;
  prop->custom_cmp = pxmutils_note_prop_cmp;
  prop->custom_free = pxmutils_note_prop_free;

  return prop;
}


/* FREE'ING FUNCTIONS
 */
gint 
pxmutils_note_free (PxmNote *note)
{
  PxmNoteVal *noteval = NULL;
  
  g_assert (note != NULL);

  if (note->name != NULL)
    g_free (note->name);
  
  /* Free the array of value strings. The array cannot be NULL, even
     if it might be empty.
   */
  g_assert (note != NULL);
  
  while (note->notevalGPA->len > 0)
    {
      noteval = g_ptr_array_remove_index (note->notevalGPA, 0);
      g_assert (noteval != NULL);
      
      pxmutils_noteval_free (noteval);
    }
  
  g_ptr_array_free (note->notevalGPA, TRUE);
  
  g_free (note);
  
  return 1;
}


gint 
pxmutils_noteval_free (PxmNoteVal *noteval)
{
  g_assert (noteval != NULL);
  
  /* Whatever the type of value, it has to be freed using g_free,
     because it was allocated using g_malloc0.
  */
  if (noteval->value != NULL)
    g_free (noteval->value);

  g_free (noteval);
  
  return 1;
}


gint 
pxmutils_noteval_GPA_empty (GPtrArray *GPA)
{
  gint count = 0;
  PxmNoteVal *noteval = NULL;
  

  g_assert (GPA != NULL);
  
  while (GPA->len > 0)
    {
      noteval = g_ptr_array_remove_index (GPA, 0);
      g_assert (noteval != NULL);
      
      pxmutils_noteval_free (noteval);

      count++;
    }

  return count;  
}


gint 
pxmutils_note_prop_free (PxmProp *prop)
{
  /* This function is called when a prop named "NOTE" must be
     freed.
   */
  g_assert (prop != NULL);
  
  if (prop->name != NULL)
    g_free (prop->name);
  
  if (prop->data != NULL)
    pxmutils_note_free ((PxmNote *) prop->data);
  
  g_free (prop);
    
  return 1;
}


gint 
pxmutils_note_prop_free_but_not_data (PxmProp *prop)
{
  /* This function is called when a prop named "NOTE" must be
     freed, but without freeing the data that were allocated in other
     places, and that are pointed to by the prop->data member.
   */
  g_assert (prop != NULL);
  
  if (prop->name != NULL)
    g_free (prop->name);
  
  g_free (prop);
    
  return 1;
}





