/* GNU polyxmass - the massist's program.
   -------------------------------------- 
   Copyright (C) 2000,2001,2002 Filippo Rusconi

   http://www.polyxmass.org

   This file is part of the "GNU polyxmass" project.
   
   The "GNU polyxmass" project is an official GNU project package (see
   www.gnu.org) released ---in its entirety--- under the GNU General
   Public License and was started at the Centre National de la
   Recherche Scientifique (FRANCE), that granted me the formal
   authorization to publish it under this Free Software License.

   This software is free software; you can redistribute it and/or
   modify it under the terms of the GNU  General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU  General Public
   License along with this software; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "pxmutils-xxxxspec.h"



/* NEW'ING FUNCTIONS, DUPLICATING FUNCTIONS ...
 */

PxmPolymerspec *
pxmutils_polymerspec_new (void)
{
  PxmPolymerspec *ps = NULL;
  
  ps = g_malloc0 (sizeof (PxmPolymerspec));
  
  ps->type = g_strdup ("");
  ps->file = g_strdup ("");
  ps->dir = g_strdup ("");
  
  return ps;
}


gboolean
pxmutils_polymerspec_set_type (PxmPolymerspec *ps, gchar *type)
{
  g_assert (ps != NULL && type != NULL);

  g_assert (ps->type != NULL);
  
  g_free (ps->type);
  
  ps->type = g_strdup (type);
  
  return TRUE;
}


gboolean
pxmutils_polymerspec_set_file (PxmPolymerspec *ps, gchar *file)
{
  g_assert (ps != NULL && file != NULL);

  g_assert (ps->file != NULL);
  
  g_free (ps->file);
  
  ps->file = g_strdup (file);
  
  return TRUE;
}


gboolean
pxmutils_polymerspec_set_dir (PxmPolymerspec *ps, gchar *dir)
{
  g_assert (ps != NULL && dir != NULL);

  g_assert (ps->dir != NULL);
  
  g_free (ps->dir);
  
  ps->dir = g_strdup (dir);
  
  return TRUE;
}






PxmMonomerspec *
pxmutils_monomerspec_new (void)
{
  PxmMonomerspec *ms = NULL;
  
  ms = g_malloc0 (sizeof (PxmMonomerspec));
    
  return ms;
}


gboolean
pxmutils_monomerspec_set_raster (PxmMonomerspec *ms, gchar *file)
{
  g_assert (ms != NULL && file != NULL);

  if (ms->raster != NULL)  
    g_free (ms->raster);
  
  ms->raster = g_strdup (file);
  
  return TRUE;
}  

gboolean
pxmutils_monomerspec_set_vector (PxmMonomerspec *ms, gchar *file)
{
  g_assert (ms != NULL && file != NULL);

  if (ms->vector != NULL)  
    g_free (ms->vector);
  
  ms->vector = g_strdup (file);
  
  return TRUE;
}  


gboolean
pxmutils_monomerspec_set_code (PxmMonomerspec *ms, gchar *code)
{
  g_assert (ms != NULL && code != NULL);

  if (ms->code != NULL)  
    g_free (ms->code);
  
  ms->code = g_strdup (code);
  
  return TRUE;
}


gboolean
pxmutils_monomerspec_set_modif (PxmMonomerspec *ms, gchar *modif)
{
  g_assert (ms != NULL && modif != NULL);

  if (ms->modif != NULL)
    g_free (ms->modif);
  
  ms->modif = g_strdup (modif);
  
  return TRUE;
}


gboolean
pxmutils_monomerspec_set_action (PxmMonomerspec *ms, gchar *action)
{
  g_assert (ms != NULL && action != NULL);

  if (ms->action != NULL)  
    g_free (ms->action);
  
  ms->action = g_strdup (action);
  
  return TRUE;
}




/*  LOCATING FUNCTIONS
 */
gint
pxmutils_polymerspec_get_index_by_type (GPtrArray *GPA, gchar *type)
{
  return pxmutils_polymerspec_get_index_top_by_type (GPA, type);
}


gint
pxmutils_polymerspec_get_index_top_by_type (GPtrArray *GPA, gchar *type)
{
  gint iter = 0;
  PxmPolymerspec *polymerspec = NULL;
  

  g_assert (GPA != NULL);
  
  for (iter = 0; iter < GPA->len; iter ++)
    {
      polymerspec = g_ptr_array_index (GPA, iter);
      
      if (0 == strcmp (polymerspec->type, type))
	return iter;
    }
  
  return -1;
}


gint
pxmutils_polymerspec_get_index_bottom_by_type (GPtrArray *GPA, gchar *type)
{
  gint iter = 0;
  PxmPolymerspec *polymerspec = NULL;
  

  g_assert (GPA != NULL);
  
  if (GPA->len > 0)
    {
      for (iter = GPA->len -1 ; iter >= 0; iter--)
	{
	  polymerspec = g_ptr_array_index (GPA, iter);
	  
	  if (0 == strcmp (polymerspec->type, type))
	    return iter;
	}
    }
    
  return -1;
}


PxmPolymerspec *
pxmutils_polymerspec_get_ptr_by_type (GPtrArray *GPA, gchar *type)
{
  return pxmutils_polymerspec_get_ptr_top_by_type (GPA, type);
}


PxmPolymerspec *
pxmutils_polymerspec_get_ptr_top_by_type (GPtrArray *GPA, gchar *type)
{
  gint iter = 0;
  PxmPolymerspec *polymerspec = NULL;
  

  g_assert (GPA != NULL);
  
  for (iter = 0; iter < GPA->len; iter ++)
    {
      polymerspec = g_ptr_array_index (GPA, iter);
      
      if (0 == strcmp (polymerspec->type, type))
	return polymerspec;
    }
  
  return NULL;
}


PxmPolymerspec *
pxmutils_polymerspec_get_ptr_bottom_by_type (GPtrArray *GPA, gchar *type)
{
  gint iter = 0;
  PxmPolymerspec *polymerspec = NULL;
  

  g_assert (GPA != NULL);
  
  if (GPA->len > 0)
    {
      for (iter = GPA->len -1 ; iter >= 0; iter--)
	{
	  polymerspec = g_ptr_array_index (GPA, iter);
	  
	  if (0 == strcmp (polymerspec->type, type))
	    return polymerspec;
	}
    }
    
  return NULL;
}


PxmMonomerspec *
pxmutils_monomerspec_get_ptr_by_code_modif (gchar *code, gchar* modif,
					    gint check, 
					    GPtrArray *GPA,
					    gint *error)
{
  gint iter = 0;
  
  PxmMonomerspec *ms = NULL;

  struct stat statbuf;

  gboolean code_found = FALSE;
  gboolean modif_found = FALSE;
  

  g_assert (code != NULL);

  g_assert (GPA != NULL);
  
  /* This function has to check several things, not only the modif name:
   *
   * 1. search for any item in the array that has its code member
   * containing (and not identical to) the code parameter.
   * This is because we are allowed this syntax "lys,arg$acetylation" for
   * example to state that both lys and arg can be acetylated...
   *
   * 2. for the items that match in 1., check that modif is non-NULL.
   * 3. If modif is non-NULL, check that the string is identical
   * to modif parameter.
   *
   * 4. check that the complying item has valid raster and vector members,
   * and that these strings point to existing files on disk, 
   * and that these filenames are ABSOLUTE.
   *
   * 5. finally, since the files seem to exist effectively, we can
   * both duplicate these file names and duplicate the action that
   * should be performed with these files (Opaque or Transparent operation).
   *
   * 6. NOTE that in the case of a rule for a modified monomer, 
   * it is allowed that the '!' be found in place of the monomer(s)'
   * codes as in "!$Phopshorylation%T%phospho.xpm", which is read
   * "Whatever the monomer code, if there is a Phosphorylation 
   * modification, perform a Transparent operation using file phospho.xpm".
   */

  for (iter = 0; iter < GPA->len; iter++)
    {
      ms = g_ptr_array_index (GPA, iter);
      
      /*
      printf ("ms at index %d is now: %s-%s-%s\n",
	      iter,
	      ms->code,
	      ms->modif,
	      ms->action);
      */

      /* Note that in the syntax of the monomer-modif.dic file, if
       * the rule is for a monomer without caring of any modif,
       * the code in this rule has to be identical to the one
       * of the monomer, while for a rule governing modified monomers,
       * degeneration is allowed by means of a comma-separated list
       * of monomer codes.
       *
       * This means that if modif is NULL, we are looking for
       * a non-modified rendering rule, which means we use
       * an identity function (strcmp ()) to search the code;
       * 
       * Further, we have to make sure that the modif member of the ms
       * instance is NULL, because we could perfectly imagine a rule
       * for rendering a modified monomer in which only one monomer is
       * specified, like
       * "S$Phosphorylation%T%phospho.svg|phospho.xpm", in which case
       * the "S" would be alone in the ms structure and if this
       * "S$Phosphorylation%T%phospho.svg|phospho.xpm" occurred before
       * the normal one "S=serine.xpm" in the monomer-modif.dic file, we
       * would return the bad ms instance.
       *
       * If the modif is non-NULL, we are looking for a rule governing
       * the rendering of one or more monomers, so we have to 
       * use a less restrictive string comparison function: strstr () and
       * clearly establish that the modif member of the ms 
       * instance is non-NULL, this time !
       */

      if (modif == NULL)
	{
	  /* We are willing to just get the ms item that has the
	     monomer code identical to the code passed as
	     parameter. Most probably because we are searching the
	     graphics file names for rendering an un-modified monomer.
	     ATTENTION: To make sure we are not selecting a ms object
	     corresponding to the other case (monomer code is
	     associated to a modif name), we make sure that ms->modif
	     is NULL.
	   */
	  if (ms->modif == NULL && 0 == strcmp (ms->code, code))
	    {
	      code_found = TRUE;
	      break;
	    }
	  else
	    {
	      continue;
	    }
	}
      
      else /* (modif != NULL) */
	{
	  /* (modif != NULL), which means we search for a monomer code
	     that is associated with a modification of name 'modif'
	     passed as parameter. Here, note that the comparison is
	     "lazy", using strstr () and not strcmp () because we
	     allow that more than one monomer code be associated with
	     a modif, like for example,
	     "Ser,Try,Thr$Phosphorylation%T%phospho.svg|phospho.xpm"
	   */
	  if (NULL != strstr (ms->code, code))
	    {
	      /* 'code' matches. Check if 'modif' matches also. Be
		 careful that we may have gotten a ms instance where
		 the modif member is NULL. So check this first.
	       */
	      code_found = TRUE;
	      
	      if (ms->modif != NULL && 0 == strcmp (ms->modif, modif))
		{
		  /* 'modif' matches. Since 'code' matched in the
		     first place, we have succeeded at our quest.
		   */
		  modif_found = TRUE;
		  break;
		}
	      else
		{
		  continue;
		}
	    }

	  /* The '!' joker character is allowed as monomer 'code',
	   * which matches all the codes from the polymer definition.
	   */
	  else if (NULL != strchr (ms->code, '!'))
	    {
	      /* Joker 'code' matches. Check if 'modif' matches
		 also. Be careful that we may have gotten a ms
		 instance where the modif member is NULL. So check
		 this first.
	       */
	      code_found = TRUE;
	      
	      if (ms->modif != NULL && 0 == strcmp (ms->modif, modif))
		{
		  /* 'modif' matches. Since 'code' matched in the
		     first place, we have succeeded at our quest.
		  */
		  modif_found = TRUE;
		  break;
		}
	      else
		{
		  continue;
		}
	    }
	  else
	    {
	      continue;
	    }
	}
    }

  /* Analyse the results. These depend on the parameter 'modif' passed
     to this function. 
  */


  /*If 'modif' is NULL, all we wanted to get is a ms for which
    ms->code == 'code', with ms->modif == NULL. Thus, in this case
    'code_found' is TRUE (first block of the for loop).
  */
  if (modif == NULL)
    {
      if (code_found == TRUE)
	{
	  if (error != NULL)
	    *error = ERR_MNMSPEC_NOERR; /* Do not return yet, still
					   checks to do.*/
	}
      
      else /* No code was found. That's an error. Return NULL
	      immediately. */
	{
	  if (error != NULL)
	    *error = ERR_MNMSPEC_NO_CODE;

	  return NULL;
	}
    }

  /*If 'modif' is non-NULL, we wanted to get a ms instance that
     matched both 'code' and 'modif' passed as parameter. Thus, in
     this case 'code_found' is TRUE and 'modif_found' must also be
     TRUE (second block of the for loop).
  */
  else /* (modif != NULL) */
    {
      if (code_found == TRUE)
	{
	  if (modif_found == TRUE)
	    {
	      if (error != NULL)
		*error = ERR_MNMSPEC_NOERR; /* Do not return yet,
					       still checks to do.*/
	    }
	  
	  else
	    {
	      if (error != NULL) /* No modif was found. That's an
				    error. Return NULL immediately. */
		*error = ERR_MNMSPEC_CODE_BUT_NO_MODIF;
	      
	      return NULL;
	    }
	}
      
      else /* No code was found. That's an error. Return NULL
	      immediately. */
	{
	  if (error != NULL)
	    *error = ERR_MNMSPEC_NO_CODE;
	      
	  return NULL;
	}
    }
  

 
  /* If we are here, that means that we found a ms that complied with
     'code' at least [and with 'code' && 'modif' if modif was
     non-NULL]. But if we found one matching ms instance and the user
     asked that a check to the graphics file not be made, return
     immediately the pointer to the found instance. Otherwise, we
     still have to work...
   */
  if (check == FILE_NONE)
    {
      /* The error code was set above.
       */
      return ms;
    }
  
  /* At this point, the user wants to check that the found ms has file
     name strings that point to existing graphics files
     (vector/raster)! There are still a number of possibilities.
   */

  /* We do not need to make an absolute filename since the file member
     of any ms instance has an absolute filename (at least this is how
     our standard is defined... catch the bugs !).
   */

  /* NOTE that FILE_BOTH means that the two graphics files MUST be
     there. If any of the two misses, then that's an error. That the
     converse of the FILE_ANY param, that is dealt with below.
  */

  /* First the vector svg file.
   */
  if (check == FILE_BOTH || check == FILE_VECTOR)
    {
      if (-1 == stat (ms->vector, &statbuf))
	{
	  /* File apparently does not exist, tell why we return NULL.
	   
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	  _("%s@%d: file was not found: %s\n"),
	  __FILE__, __LINE__, ms->vector);
	  */

	  if (error != NULL)
	    *error = ERR_MNMSPEC_NO_VECTOR_FILE;
	  
	  return NULL;
	}
    }
  
  /* Second the raster svg file.
   */
  if (check == FILE_BOTH || check == FILE_RASTER)
    {
      if (-1 == stat (ms->raster, &statbuf))
	{
	  /* File apparently does not exist, tell why we return NULL.
	     
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	  _("%s@%d: file was not found: %s\n"),
	  __FILE__, __LINE__, ms->raster);
	  */
	  
	  if (error != NULL)
	    *error = ERR_MNMSPEC_NO_RASTER_FILE;
	  
	  return NULL;
	}
    }

  /* Finally the special case FILE_ANY. This means that we do not care
     if graphics file available is vector or raster. All we want is
     that ONE be there. If both are absent, then that's an error.
   */
  if (check == FILE_ANY)
    {
      if (-1 == stat (ms->raster, &statbuf) 
	  && -1 == stat (ms->raster, &statbuf))
	{
	  /* File apparently does not exist, tell why we return NULL.
	   */
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 _("%s@%d: either one or both of files was not found: "
		   "%s,%s\n"),
		 __FILE__, __LINE__, ms->vector, ms->raster);
	  
	  if (error != NULL)
	    *error = ERR_MNMSPEC_NO_VECTOR_NOR_RASTER_FILES;
	  
	  return NULL;
	}
    }
  
  
  /* Apparently the item that we found in array contains valid data,
   * we just return a pointer to it.
   */
  return ms;
}


gint
pxmutils_monomerspec_get_index_by_code (GPtrArray *GPA, gchar *code)
{
  return pxmutils_monomerspec_get_index_top_by_code (GPA, code);
}


gint
pxmutils_monomerspec_get_index_top_by_code (GPtrArray *GPA, gchar *code)
{
  gint iter = 0;
  PxmMonomerspec *monomerspec = NULL;
  

  g_assert (GPA != NULL);
  
  for (iter = 0; iter < GPA->len; iter ++)
    {
      monomerspec = g_ptr_array_index (GPA, iter);
      
      if (0 == strcmp (monomerspec->code, code))
	return iter;
    }
  
  return -1;
}


gint
pxmutils_monomerspec_get_index_bottom_by_code (GPtrArray *GPA, gchar *code)
{
  gint iter = 0;
  PxmMonomerspec *monomerspec = NULL;
  

  g_assert (GPA != NULL);
  
  if (GPA->len > 0)
    {
      for (iter = GPA->len -1 ; iter >= 0; iter--)
	{
	  monomerspec = g_ptr_array_index (GPA, iter);
	  
	  if (0 == strcmp (monomerspec->code, code))
	    return iter;
	}
    }
    
  return -1;
}


gint
pxmutils_monomerspec_get_index_by_modif (GPtrArray *GPA, gchar *modif)
{
  return pxmutils_monomerspec_get_index_top_by_modif (GPA, modif);
}


gint
pxmutils_monomerspec_get_index_top_by_modif (GPtrArray *GPA, gchar *modif)
{
  gint iter = 0;
  PxmMonomerspec *monomerspec = NULL;
  

  g_assert (GPA != NULL);
  
  for (iter = 0; iter < GPA->len; iter ++)
    {
      monomerspec = g_ptr_array_index (GPA, iter);
      
      if (0 == strcmp (monomerspec->modif, modif))
	return iter;
    }
  
  return -1;
}


gint
pxmutils_monomerspec_get_index_bottom_by_modif (GPtrArray *GPA, gchar *modif)
{
  gint iter = 0;
  PxmMonomerspec *monomerspec = NULL;
  

  g_assert (GPA != NULL);
  
  if (GPA->len > 0)
    {
      for (iter = GPA->len -1 ; iter >= 0; iter--)
	{
	  monomerspec = g_ptr_array_index (GPA, iter);
	  
	  if (0 == strcmp (monomerspec->modif, modif))
	    return iter;
	}
    }
    
  return -1;
}




/* UTILITY FUNCTIONS
 */
gboolean
pxmutils_polymerspec_reset (PxmPolymerspec *ps)
{
  g_assert (ps != NULL);
  
  g_assert (ps->type != NULL);
  g_free (ps->type);
  ps->type = g_strdup ("");
  
  g_assert (ps->file != NULL);
  g_free (ps->file);
  ps->file = g_strdup ("");
  
  g_assert (ps->dir != NULL);
  g_free (ps->dir);
  ps->dir = g_strdup ("");
  
  return TRUE;
}


gint
pxmutils_polymerspec_fill_array_from_dict (gchar *file,
					   GPtrArray *GPA)
{
  gint count = 0;
  
  gchar *line = NULL;
  gchar *read = NULL;
  gchar *dir = NULL;

  FILE *stream = NULL;

  PxmPolymerspec *ps = NULL;

  g_assert (GPA != NULL);
  g_assert (file != NULL);
  
  if (strlen (file) <= 0)
    return -1;
  
  /* Note that we NEED that the file name passed as parameter be 
   * an absolute file name because we are going to use the dirname
   * to construct the PxmPolymerspec data.
   */
  if (FALSE == g_path_is_absolute (file))
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: file name is not absolute: '%s'\n"),
	     __FILE__, __LINE__, file);

      return -1;
    }

  /* Get the dir name from the abolute file name.
   */
  dir = g_path_get_dirname (file);

  stream = fopen (file, "r");
  
  if (NULL == stream)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: file not found: %s\n"),
	     __FILE__, __LINE__, file);
      
      return -1;
    }

  /* The lines that we have to parse here are in this format:
   * polname=polname.xml%poldir, which translates into
   * protein=protein.xml%protein in the case of the "protein" polymer
   * definition.
   */

  /* Allocate the memory into which each line of the file will be
   * store for recursive parsing.
   */
  line = g_malloc0 (sizeof (gchar) * (MAX_LINE_LENGTH + 1));

  read = line;

  while (read != NULL)
    {
      read = fgets (read, MAX_LINE_LENGTH, stream);

      if (read != NULL && strlen (read) > 1)
	{
	  /* Apparently something interesting should be in 'read' for
	     us to parse.
	  */
	  /* We first have to check if the line we are reading is
	     actually a comment line, beginning with the '#'
	     character. The same applies if the line is empty (starts
	     with a newline...
	   */
	  if (read [0] == '#' || read [0] == '\n')
	    continue;

	  while (read [0] != '\x0')
	    {
	      if (TRUE == g_ascii_isspace (read [0]))
		read++;
	      else
		break;
	    }
	  if (strlen (read) <= 0)
	    continue;
	  
	  /* Next check if we are on a line that describes the
	   * correspondence between a polymer definition name and its
	   * definition file and the pixmaps directory. This is most
	   * probably the case if both '=' and '%' are present in the
	   * line.
	   */
	  if (NULL != strchr (read, '=') 
	      && NULL != strchr (read, '%'))
	    {
	      /* We are dealing with a 
	       * polname=poldef.dat%poldir correspondence.
	       */
	      
	      /* Allocate the PxmPolymerspec instance that we'll
	       * initialize using the data read from parsing of the
	       * line.
	       */
	      ps = pxmutils_polymerspec_new ();
	      
	      if (FALSE == pxmutils_polymerspec_init_from_line (read,
								dir,
								ps))
		{
		  g_free (line);
		  
		  pxmutils_polymerspec_free (ps);
		  
		  fclose (stream);
		  
		  return -1;
		}
	      
	      g_ptr_array_add (GPA, ps);
	      
	      /*
		debug_printf (("added new ps: %s--%s--%s\n",
		ps->type, ps->file, ps->dir));
	      */
	      
	      count++;

	      continue;
	    }
	}
    }

  /* Now that we do not need this string anymore, since the file is
   * finished parsing, we can free the memory for reuse of the pointer.
   */
  g_free (line);

  fclose (stream);

  return count;
}


gboolean
pxmutils_polymerspec_init_from_line (gchar *line,
				     gchar *dir,
				     PxmPolymerspec *ps) 
{
  gint iter = 0;

  GString * gs = NULL;

  gboolean type_done = FALSE;
  gboolean file_done = FALSE;

  gchar *path = NULL;
  
  g_assert (ps != NULL);

  gs = g_string_new ("");

  /* We have a line of the kind
   * "type=poldef.xml%/absolute/path/pixmapsdir" (note the newline
   * character may be found if the line in question is not the last of
   * the file!), we just need to deconstruct this into three pieces.
   */
  while (line [iter] != '\x0' && line [iter] != '\n')
    {
      if (line [iter] == '=')
	{
	  /* We should now have the polymer definition name in the gs
	   * GString.
	   */
	  if (strlen (gs->str) <= 0)
	    {
	      g_string_free (gs, TRUE);

	      return FALSE;
	    }

	  pxmutils_polymerspec_set_type (ps, gs->str);
	  
	  gs = g_string_truncate (gs, 0);

	  /*
	    g_string_free (gs, TRUE);
	    gs= g_string_new ("");
	  */
	  
	  type_done = TRUE;
	}

      else if (line [iter] == '%')
	{
	  /* We should now have the polymer definition file name in
	   * the gs GString.
	   */
	  if (strlen (gs->str) <= 0)
	    {
	      g_string_free (gs, TRUE);
	      
	      return FALSE;
	    }

	  /* Remember that what we get is a file name that is relative
	   * to the polyxmass polymer definition configuration directory.
	   * Thus we need to "absolutize" the filename prior to setting
	   * it to the polymerspec file member.
	   */
	  path = g_strdup_printf ("%s/%s", dir, gs->str); 
	  
	  pxmutils_polymerspec_set_file (ps, path);
	  
	  g_free (path);
	  
	  gs = g_string_truncate (gs, 0);

	  /*
	    g_string_free (gs, TRUE);
	    gs= g_string_new ("");
	  */

	  file_done = TRUE;
	}

      else
	{
	  /* We are iterating in some character, so we should simply
	   * put this into the gs GString.
	   */
	  gs = g_string_append_c (gs, line [iter]);
	}

      iter++;		    
    }

  /* And now we are at the end of the line, so the directory should
   * have completed now, but check that the type and file were done
   * already.
   */
  if (type_done == FALSE || file_done == FALSE)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: error parsing a polymerspec line: '%s'\n"),
	     __FILE__, __LINE__, line);
      
      g_string_free (gs, TRUE);
      
      return FALSE;
    }

  /* We should now have the directory in the gs GString.
   */
  /* Remember that what we get is a dir name that is relative
   * to the polyxmass polymer definition configuration directory.
   * Thus we need to "absolutize" the filename prior to setting
   * it to the polymerspec file member.
   */
  path = g_strdup_printf ("%s/%s", dir, gs->str); 
  
  pxmutils_polymerspec_set_dir (ps, path);
  
  g_free (path);

  /* We are no longer going to use this allocated GString.
   */
  g_string_free (gs, TRUE);
  
  return TRUE;
}


gboolean
pxmutils_polymerspec_init_from_dict_with_type (gchar *type,
					       gchar *file,
					       PxmPolymerspec *ps) 
{
  gchar *line = NULL;
  gchar *read = NULL;
  gchar *dir = NULL;

  FILE *stream = NULL;

  g_assert (type && strlen (type) > 0);
  g_assert (ps != NULL);

  g_assert (file != NULL);
  
  if (strlen (file) <= 0)
    return FALSE;

  /* Note that we NEED that the file name passed as parameter be 
   * an absolute file name because we are going to use the dirname
   * to construct the PxmPolymerspec data.
   */
  if (FALSE == g_path_is_absolute (file))
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: file name is not absolute: '%s'\n"),
	     __FILE__, __LINE__, file);

      return FALSE;
    }

  /* Get the dir name from the abolute file name.
   */
  dir = g_path_get_dirname (file);

  stream = fopen (file, "r");
  
  if (NULL == stream)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: file not found: %s\n"),
	     __FILE__, __LINE__, file);
      
      g_free (dir);
      
      return FALSE;
    }

  /* The lines that we have to parse here are in this format:
   * polname=polname.xml%poldir, which translates into
   * protein=protein.xml%protein in the case of the "protein" polymer
   * definition.
   */

  /* Allocate the memory into which each line of the file will be
   * store for recursive parsing.
   */
  line = g_malloc0 (sizeof (gchar) * (MAX_LINE_LENGTH + 1));

  read = line;

  while (read != NULL)
    {
      read = fgets (read, MAX_LINE_LENGTH, stream);

      if (read != NULL && strlen (read) > 1)
	{
	  /* Apparently something interesting should be 
	   * in 'read' for us to parse.
	   */
	  /* We first have to check if the line we are reading is
	   * actually a comment line, beginning with the '#'
	   * character.
	   */
	  if (read [0] == '#')
	    continue;
	  
	  /* Next check if we are on a line that describes the
	   * correspondence between a polymer definition name and its
	   * definition file and the pixmaps directory. This is most
	   * probably the case if both '=' and '%' are present in the
	   * line.
	   */
	  if (NULL != strchr (read, '=') 
	      && NULL != strchr (read, '%'))
	    {
	      /* We are dealing with a polname=poldef.dat%poldir
		 correspondence.
	       */
	      if (TRUE == pxmutils_polymerspec_init_from_line (read,
							       dir,
							       ps))
		{
		  /* We got a PxmPolymerspec back, so check its type 
		   * against the one we're looking for. If the type is
		   * OK, just return TRUE, so that caller knows
		   * that the ps structure now contains correct data.
		   */
		  if (0 == strcmp (ps->type, type))
		    {
		      g_free (line);
		      
		      fclose (stream);

		      g_free (dir);
		      
		      return TRUE;
		    }
		  else
		    {
		      pxmutils_polymerspec_reset (ps);
		      
		      continue;
		    }
		}
	      else
		{
		  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
			 _("%s@%d: error parsing a polymerspec"
			   " line: '%s'\n"),
			 __FILE__, __LINE__, read);

		  g_free (line);	      
		  
		  fclose (stream);

		  g_free (dir);
		  
		  return FALSE;
		  
		}
	    }
	}
    }
  
  g_free (line);
  
  g_free (dir);
  
  fclose (stream);
  
  return FALSE;
}


gint
pxmutils_monomerspec_fill_array_from_dict (gchar *file,
					   GPtrArray *GPA)
{
  gint count = 0;

  gchar *line = NULL; 
  gchar *read = NULL;
  gchar * dir = NULL;

  FILE *stream = NULL;

  PxmMonomerspec *ms = NULL;

  g_assert (GPA != NULL);
  g_assert (file != NULL);
  
  if (strlen (file) <= 0)
    return -1;
  
  /* Note that we NEED that the file name passed as parameter be 
   * an absolute file name because we are going to use the dirname
   * to construct the PxmPolymerspec data.
   */
  if (FALSE == g_path_is_absolute (file))
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: file name is not absolute: '%s'\n"),
	     __FILE__, __LINE__, file);

      return -1;
    }

  /* Get the dir name from the abolute file name.
   */
  dir = g_path_get_dirname (file);

  stream = fopen (file, "r");
  
  if (NULL == stream)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: file not found: %s\n"),
	     __FILE__, __LINE__, file);

      g_free (dir);
      
      return -1;
    }

  /* The lines that we have to parse here are in these two formats:
   * 1. R=arginine.svg|arginine.xpm
   * 2. D$Amidation%O%asparagine.xpm
   *
   * Note that one of the two graphics files may miss. Which is that
   * we may have something like:
   *
   * R=arginine.svg|
   * or
   * R=|arginine.xpm
   */

  /* Allocate the memory into which each line of the file will be
   * store for recursive parsing.
   */
  line = g_malloc0 (sizeof (gchar) * (MAX_LINE_LENGTH + 1));

  read = line;

  while (read != NULL)
    {
      read = fgets (read, MAX_LINE_LENGTH, stream);

      if (read != NULL && strlen (read) > 1)
	{
	  /* Apparently something interesting should be 
	   * in 'read' for us to parse.
	   */
	  /* We first have to check if the line we are reading is
	   * actually a comment line, beginning with the '#'
	   * character. The same applies if the line is empty (starts
	     with a newline...
	   */
	  if (read [0] == '#' || read [0] == '\n')
	    continue;

	  while (read [0] != '\x0')
	    {
	      if (TRUE == g_ascii_isspace (read [0]))
		read++;
	      else
		break;
	    }
	  if (strlen (read) <= 0)
	    continue;
	  
	  /* Next check if we are on a line that describes the
	   * correspondence between a polymer definition name and its
	   * definition file and the pixmaps directory. This is most
	   * probably the case if both '=' and '|' are present in the
	   * line.
	   */
	  if (NULL != strchr (read, '=') 
	      && NULL != strchr (read, '|'))
	    {
	      /* We are dealing with a monomcode=pixmap.svg|pixmap.png
		 correspondence.
	       */
	      
	      ms = pxmutils_monomerspec_parse_mnm_line (read, dir);
	      
	      if (NULL == ms)
		{
		  g_free (line);
		  
		  fclose (stream);

		  g_free (dir);
		  
		  return -1;
		}
	      else
		{
		  count++;
		  
		  g_ptr_array_add (GPA, ms);
		  
		  continue;
		}
	    }
	  
	  /* The format of the file is such that if there is no '='
	   * character, then there is necessarily a '$' character.
	   * Otherwise, the file is not well-formed.
	   */
	  else if (NULL != strchr (read, '$'))
	    {
	      /* Apparently we are dealing with a modif 
	       * correspondence line, which may be of the forms:
	       *
	       * S$Phosphorylation%T%phospho.xpm
	       * Y$Phosphorylation%T%phospho.svg|phospho.xpm
	       * D$Amidation%O%asparagine.svg|asparagine.xpm
	       * !$Acetylation%T%acetyl.svg|acetyl.xpm
	       * E$Amidation%O%glutamine.svg|glutamine.xpm
	       */
	      ms = pxmutils_monomerspec_parse_mdf_line (read, dir);

	      if (NULL == ms)
		{
		  g_free (line);
		  
		  fclose (stream);

		  g_free (dir);
		  
		  return -1;
		}
	      else
		{
		  count++;

		  g_ptr_array_add (GPA, ms);

		  continue;
		}
	    }
	  
	  else
	    {
	      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		     _("%s@%d: file not well formed: %s\n"),
		     __FILE__, __LINE__, file);
	      
	      g_free (line);

	      pxmutils_monomerspec_free (ms);
	      
	      fclose (stream);

	      g_free (dir);
	      
	      return -1;
	    }
	}
    }
  /*end of   while (1)
   * parsing file's lines.
   */

  /* Now that we do not need this string anymore, since the file is
   * finished parsing, we can free the memory for reuse of the pointer
   */
  g_free (line);
  
  fclose (stream);

  g_free (dir);
  
  return count;
}


PxmMonomerspec *
pxmutils_monomerspec_parse_mnm_line (gchar *line, gchar *dir) 
{
  gint iter = 0;

  GString * gs = NULL;

  gchar *help = NULL;
  
  gboolean code_done = FALSE;

  gboolean vector_done = FALSE;
  gboolean raster_done = FALSE;
  
  PxmMonomerspec *ms = NULL;


  g_assert (line != NULL);
  g_assert (dir != NULL);
  
  gs = g_string_new ("");
  g_assert (gs != NULL);

  /* We have a line of the kind "code=pixmapfile.svg|pixmapfile.xpm"
   * (note the newline character may be found if the line in question
   * is not the last of the file!), we just need to deconstruct this
   * into three pieces: code, pixmapfile.svg and pixmapfile.xpm.
   */

  /* The lines that we have to parse here are in these two formats:
   * 1. R=arginine.svg|arginine.xpm
   * 2. D$Amidation%O%asparagine.xpm
   *
   * Note that one of the two graphics files may miss. Which is that
   * we may have something like:
   *
   * R=arginine.svg|
   * or
   * R=|arginine.xpm
   */

  /* Allocate the monomerspec instance that we'll fill with data 
   * parsed out of the currently analyzed line.
   */
  ms = pxmutils_monomerspec_new ();

  while (line [iter] != '\x0' && line [iter] != '\n')
    {
      if (line [iter] == '=')
	{
	  /* We should have something in gs, namely the code.
	   */
	  if (strlen (gs->str) <= 0)
	    {
	      g_string_free (gs, TRUE);
	      
	      pxmutils_monomerspec_free (ms);
	      
	      return NULL;
	    }

	  pxmutils_monomerspec_set_code (ms, gs->str);
	  
	  gs = g_string_truncate (gs, 0);
	  
	  code_done = TRUE;
	}

      else if (line [iter] == '|')
	{
	  /* We should have something in gs, namely the vector
	     filename. However, the string gs may be empty, as
	     explained above.
	   */
	  if (strlen (gs->str) > 0)
	    {
	      /* Make an absolute filename.
	       */
	      help = g_strdup_printf ("%s/%s", dir, gs->str);
	      
	      pxmutils_monomerspec_set_vector (ms, help);
	      
	      g_free (help);
	      
	      gs = g_string_truncate (gs, 0);
	      
	      vector_done = TRUE;
	    }
	}
      
      else
	{
	  /* We are iterating in some character, so we should simply
	   * put this into the gs string.
	   */
	  gs = g_string_append_c (gs, line[iter]);
	}

      iter++;		    
    }

  /* And now we are at the end of the line, so the raster file name
   * should have completed now, in gs as above. However, the gs string
   * may be empty, because the raster filename may be empty, if the
   * vector filename is not. So we have to check this now. Also, do
   * not forget to check that the code was done already.
   */
  if (strlen (gs->str) > 0)
    {
      /* Make an absolute filename.
       */
      help = g_strdup_printf ("%s/%s", dir, gs->str);
      
      pxmutils_monomerspec_set_raster (ms, help);
      
      g_free (help);

      raster_done = TRUE;
    }

  else 
    {
      if (vector_done == FALSE)
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 _("%s@%d: both vector/raster file names cannot be emtpy\n"),
		 __FILE__, __LINE__);

	  pxmutils_monomerspec_free (ms);
	  
	  g_string_free (gs, TRUE);
	  
	  return NULL;
	}
    }
  
  /* Apparently, at least one of the two vector/raster file names 
   * is not empty, which is good. We still have to check that the
   * code was set properly in a previous parsing step.
   */
  if (code_done == FALSE)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: line not well formed: %s\n"),
	     __FILE__, __LINE__, line);
      
      pxmutils_monomerspec_free (ms);
      
      g_string_free (gs, TRUE);
      
      return NULL;
    }
  
  g_string_free (gs, TRUE);
  
  return ms;
}


					    

PxmMonomerspec *
pxmutils_monomerspec_parse_mdf_line (gchar *line, gchar *dir) 
{
  gint iter = 0;

  GString * gs = NULL;

  gchar *help = NULL;

  gboolean code_done = FALSE;
  gboolean modif_done = FALSE;
  gboolean action_done = FALSE;

  gboolean vector_done = FALSE;
  gboolean raster_done = FALSE;

  PxmMonomerspec *ms = NULL;


  g_assert (line != NULL);
  g_assert (dir != NULL);


  gs = g_string_new ("");
  g_assert (gs != NULL);

  /* We have a line of the kind
   * "Ser,Tyr$Phosphorylation%T%phospho.svg|phospho.xpm" (note the
   * newline character may be found if the line in question is not the
   * last of the file!), we just need to deconstruct this into pieces.
   *
   * Note that one of the two graphics files may miss. Which is that
   * we may have something like:
   *
   * "D$Amidation%O%asparagine.svg|asparagine.xpm"
   * or
   * "Ser,Tyr$Phosphorylation%T%|phospho.xpm"
   * or
   * "Ser,Tyr$Phosphorylation%T%phospho.svg|"
   */

  /* Allocate the monomerspec instance that we'll fill with data 
   * parsed out of the currently analyzed line.
   */
  ms = pxmutils_monomerspec_new ();

  while (line [iter] != '\x0' && line [iter] != '\n')
    {
      if (line [iter] == '$')
	{
	  /* Now that we have reached the '$' character, 
	   * we should have something in gs, namely the monomer code.
	   */
	  if (strlen (gs->str) <= 0)
	    {
	      g_string_free (gs, TRUE);

	      pxmutils_monomerspec_free (ms);

	      return NULL;
	    }

	  /* We do not check for code validity because in our present
	   * case the gs->str string may contain a complex pattern
	   * including a '!' joker character...
	   */
	  pxmutils_monomerspec_set_code (ms, gs->str);

	  gs = g_string_truncate (gs, 0);
	  
	  code_done = TRUE;
	}

      else if (line [iter] == '%')
	{
	  /* There are two '%' signs in a line. When we reach the
	   * first, that means that we have just finished parsing the
	   * modif name, which is modif_done is FALSE. When we reach
	   * the second, that means that we already have finished
	   * parsing and validating the modification name, which is
	   * the name_one is TRUE. Just work out these two
	   * possiblities, depending on the value of modif_done.
	   */
	  if (modif_done == FALSE)
	    {
	      /* gs should contain the modif name just terminated 
	       * parsing.
	       */
	      pxmutils_monomerspec_set_modif (ms, gs->str);

	      gs = g_string_truncate (gs, 0);

	      modif_done = TRUE;
	    }
	  else
	    {
	      /* gs should contain the action type string: "T" for 
	       * transparent, and "O" for opaque.
	       */
	      if (0 != strcmp (gs->str, "T") 
		  && 0 != strcmp (gs->str, "O"))
		{
		  g_string_free (gs, TRUE);
		 
		  pxmutils_monomerspec_free (ms);
		  
		  return NULL;
		}

	      pxmutils_monomerspec_set_action (ms, gs->str);
	  
	      gs = g_string_truncate (gs, 0);

	      action_done = TRUE;
	    }
	}
      
      else if (line [iter] == '|')
	{
	  /* We should have something in gs, namely the vector
	     filename. However, the string gs may be empty, as
	     explained above.
	   */
	  if (strlen (gs->str) > 0)
	    {
	      /* Make an absolute filename.
	       */
	      help = g_strdup_printf ("%s/%s", dir, gs->str);
      
	      pxmutils_monomerspec_set_vector (ms, help);

	      g_free (help);
	      
	      gs = g_string_truncate (gs, 0);
	      
	      vector_done = TRUE;
	    }
	}

      else
	{
	  /* we are iterating in some character, so we should simply
	   * put this into the gs string.
	   */
	  gs = g_string_append_c (gs, line[iter]);
	}
      
      iter++;		    
    }

  /* And now we are at the end of the line, so the raster file name
   * should have completed now, in gs as above. However, the gs string
   * may be empty, because the raster filename may be empty, if the
   * vector filename is not. So we have to check this now. Also, do
   * not forget to check that the code was done already.
   */
  if (strlen (gs->str) > 0)
    {
      /* Make an absolute filename.
       */
      help = g_strdup_printf ("%s/%s", dir, gs->str);
      
      pxmutils_monomerspec_set_raster (ms, help);
      
      g_free (help);
      
      raster_done = TRUE;
    }

  else 
    {
      if (vector_done == FALSE)
	{
	  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
		 _("%s@%d: both vector/raster file names cannot be emtpy\n"),
		 __FILE__, __LINE__);

	  pxmutils_monomerspec_free (ms);
	  
	  g_string_free (gs, TRUE);
	  
	  return NULL;
	}
    }
  
  /* And now we are at the end of the line, so the file name should
   * have completed now, but check that the other elements were done already.
   */
  if (code_done == FALSE 
      || modif_done == FALSE 
      || action_done == FALSE)
    {
      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
	     _("%s@%d: line not well formed: %s\n"),
	     __FILE__, __LINE__, line);
      
      pxmutils_monomerspec_free (ms);

      g_string_free (gs, TRUE);
      
      return NULL;
    }

  g_string_free (gs, TRUE);
  
  return ms;
}



/* FREE'ING FUNCTIONS
 */
gint
pxmutils_polymerspec_free (PxmPolymerspec *ps)
{
  g_assert (ps != NULL);
  
  if (ps->type != NULL)
    g_free (ps->type);
  
  if (ps->file != NULL)
    g_free (ps->file);
  
  if (ps->dir != NULL)
    g_free (ps->dir);
  
  g_free (ps);
  
  return 1;
}


gint
pxmutils_monomerspec_free (PxmMonomerspec *ms)
{
  g_assert (ms != NULL);
  
  if (ms->raster != NULL)
    g_free (ms->raster);
  
  if (ms->vector != NULL)
    g_free (ms->vector);
  
  if (ms->code != NULL)
    g_free (ms->code);
  
  if (ms->modif != NULL)
    g_free (ms->modif);
  
  if (ms->action != NULL)
    g_free (ms->action);
  
  g_free (ms);
  
  return 1;
}




/* GPtrArray-RELATED FUNCTIONS
 */
gint
pxmutils_polymerspec_GPA_free (GPtrArray *GPA)
{
  gint count = 0;
  
  PxmPolymerspec *ps = NULL;
  

  g_assert (GPA != NULL);
  
  while (GPA->len > 0)
    {
      ps = g_ptr_array_remove_index (GPA, 0);
      g_assert (ps != NULL);
      pxmutils_polymerspec_free (ps);
      count++;
    }
  
  g_ptr_array_free (GPA, TRUE);

  return count;
}


gint
pxmutils_monomerspec_GPA_empty (GPtrArray *GPA)
{
  gint count = 0;
  
  PxmMonomerspec *ms = NULL;
  

  g_assert (GPA != NULL);
  
  while (GPA->len > 0)
    {
      ms = g_ptr_array_remove_index (GPA, 0);
      g_assert (ms != NULL);
      pxmutils_monomerspec_free (ms);
      count++;
    }
  
  return count;
}


gint
pxmutils_monomerspec_GPA_free (GPtrArray *GPA)
{
  gint count = 0;
  

  g_assert (GPA != NULL);

  count = pxmutils_monomerspec_GPA_empty (GPA);
    
  g_ptr_array_free (GPA, TRUE);

  return count;
}


